/*------------------------------------------------------------------------
Apo to Biblio:
I.Vlahavas, P.Kefalas, N.Basileiadis, F.Kokkoras, I.Sakellariou
Texnhth Nohmosunh (B' Ekdosh), Ekdoseis Gartaganh, Thessaloniki, 2005
http://aibook.csd.auth.gr
------------------------------------------------------------------------*/

/*------------------------------------------------------------
  BEAM SEARCH
	Prompt for the size of the Frontier.
	If Size=1 then it is Hill Climbing
-------------------------------------------------------------*/

gobeam(Solution):-
	write('What is the size of frontier you wish to have? '),
	read(FSize),		% This must be an integer >=1.
	initial_state(IS),
	heuristic(IS,V),
	bestfs([V-[IS]],Solution, FSize).

bestfs([_-[State|Path]|_],[State|Path],_):-
	goal(State).
bestfs([V-[State|Path]|RestPaths],Solution,FSize):-
	nl, write([V-[State|Path]|RestPaths]), 
%	nl, write('Expanding best state :'), write(V-State),nl,
	expand(V-[State|Path],NewPaths),
	append(NewPaths,RestPaths,Frontier),
	keysort(Frontier,OrderedFrontier),
	arrange_beam(OrderedFrontier, FSize, BeamFrontier),
	bestfs(BeamFrontier,Solution,FSize).
			 
expand(_-[State|Path],NewPaths):-
	findall( Value-[NewState,State|Path],
			(
			operator(State,NewState),
			not(member(NewState,Path)),
			heuristic(NewState,Value)
			),
		NewPaths).

heuristic(State,V):-
	findall(Value, 
			(goal(GoalState),h(State,GoalState,Value)), 
	        AllValues),
	min(AllValues,V),!.
	
min([M],M):-!.
min([H|T],H):-
	min(T,M),
	H=<M.
min([H|T],M):-
	min(T,M),
	H>M.

arrange_beam([], _, []):-!.
arrange_beam([H|_], 1, [H]):-!.
arrange_beam([H|T], FSize, [H|R]):-
	NSize is FSize-1,
	arrange_beam(T, NSize, R),!.

